from contextlib import contextmanager
from typing import Callable

from sqlalchemy.exc import OperationalError

from petisco.persistence.sqlalchemy.sqlalchemy_operational_database_error import (
    SqlAlchemyOperationalDatabaseError,
)


def sql_session_scope_provider(Session) -> Callable:
    @contextmanager
    def session_scope():
        session = Session()
        try:
            yield session
            session.commit()
        except OperationalError as e:
            print(e)
            session.rollback()
            raise SqlAlchemyOperationalDatabaseError
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()
        Session.remove()

    return session_scope



from typing import Dict

from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import StreamLostError
from pika import BasicProperties

from petisco.event.shared.domain.event import Event
from petisco.event.legacy.publisher.domain.interface_event_publisher import (
    IEventPublisher,
)

from petisco.event.legacy.rabbitmq.create_exchange_and_bind_queue import (
    create_exchange_and_bind_queue,
    create_dead_letter_exchange_and_bind_queue,
)
from petisco.event.legacy.rabbitmq.get_event_binding_key import get_event_binding_key
from petisco.event.shared.infrastructure.rabbitmq.rabbitmq_connector import (
    RabbitMqConnector,
)


class RabbitMQEventPublisher(IEventPublisher):
    def __init__(
        self, connector: RabbitMqConnector, organization: str, service: str, topic: str
    ):
        self.connector = connector
        self.organization = organization
        self.exchange = service
        self.queue = topic
        self.binding_key = get_event_binding_key(organization, service)
        self.properties = self._get_message_persistent_properties()
        self._connect()
        channel = self._get_channel()
        self._setup_exchanges_and_queues(channel)
        super().__init__()

    def _connect(self):
        if not self.connector:
            raise TypeError(f"RabbitMQEventPublisher: Invalid Given RabbitMQConnector")
        self.connection = self.connector.get_connection(
            f"{self.organization}.{self.exchange}"
        )

    def _setup_exchanges_and_queues(self, channel: BlockingChannel):
        create_dead_letter_exchange_and_bind_queue(
            channel=channel,
            exchange=self.exchange,
            queue=self.queue,
            binding_key=self.binding_key,
        )
        create_exchange_and_bind_queue(
            channel=channel,
            exchange=self.exchange,
            queue=self.queue,
            binding_key=self.binding_key,
            dead_letter=True,
        )

    def _get_event_routing_key(self, event: Event):
        """
        acme.onboarding.1.event.user.created
          |       |     |        |      |-> action (past verb)
          |       |     |        |-> domain entity
          |       |     |-> version
          |       |-> service
          |-> organization
        """
        return f"{self.organization}.{self.exchange}.{event.event_version}.event.{event.event_name}"

    def _get_message_persistent_properties(self):
        """
        Make message persistent (PERSISTENT_TEXT_PLAIN)
        """
        return BasicProperties(delivery_mode=2)

    def info(self) -> Dict:
        return {
            "name": self.__class__.__name__,
            "connection.is_open": self.connection.is_open,
        }

    def close(self):
        if self.connection.is_open:
            self.connection.close()

    def _check_connection(self):
        if not self.connection.is_open:
            self._connect()

    def _get_channel(self) -> BlockingChannel:
        self._check_connection()
        try:
            channel = self.connection.channel()
        except StreamLostError:
            self._check_connection()
            channel = self.connection.channel()
        return channel

    def publish(self, event: Event):

        if not event:
            return

        channel = self._get_channel()

        routing_key = self._get_event_routing_key(event)

        channel.basic_publish(
            exchange=self.exchange,
            routing_key=routing_key,
            body=event.to_json(),
            properties=self.properties,
        )

        channel.close()



from petisco import InfoId, ClientId, UserId
from petisco.domain.aggregate_roots.aggregate_root import AggregateRoot
from petisco.domain.value_objects.name import Name
from tests.end2end.flask_app.user.domain.events.user_created import UserCreated


class User(AggregateRoot):
    def __init__(self, name: Name, client_id: ClientId, user_id: UserId):
        self.client_id = client_id
        self.user_id = user_id
        self.name = name
        super().__init__()

    @staticmethod
    def create(info_id: InfoId, name: Name):
        user = User(name, info_id.client_id, info_id.user_id)
        user_created = UserCreated(name).add_info_id(info_id)
        user.record(user_created)
        return user

    def to_dict(self):
        return {
            "user_id": self.user_id.value,
            "client_id": self.client_id.value,
            "name": self.name.value,
        }



import traceback

import pytest
from meiga.assertions import assert_failure
from meiga.decorators import meiga

from petisco.domain.errors.critical_error import CriticalError


@pytest.mark.unit
def test_should_construct_critical_error_from_base_exception():

    error = CriticalError(Exception("my_trace"))

    assert error.message == "Exception: my_trace"


@pytest.mark.unit
def test_should_construct_critical_error_from_runtime_error_exception():

    error = CriticalError(RuntimeError("my_trace"))

    assert error.message == "RuntimeError: my_trace"


@pytest.mark.unit
def test_should_capture_an_exception_during_runtime():

    expected_raised_exception = RuntimeError("my_trace")

    def inner():
        raise expected_raised_exception

    @meiga
    def method():
        try:
            inner()
        except Exception as exception:
            raise CriticalError(exception)

    result = method()

    assert_failure(
        result,
        value_is_instance_of=CriticalError,
        value_is_equal_to=CriticalError(expected_raised_exception),
    )


@pytest.mark.unit
def test_should_capture_an_exception_with_additional_context():

    expected_trace = "my_trace"
    expected_raised_exception = RuntimeError(expected_trace)

    def inner():
        raise expected_raised_exception

    @meiga
    def method():
        try:
            inner()
        except Exception as exception:
            raise CriticalError(
                exception=exception,
                executor="executor",
                traceback=traceback.format_exc(),
            )

    result = method()
    assert_failure(
        result,
        value_is_instance_of=CriticalError,
        value_is_equal_to=CriticalError(expected_raised_exception),
    )

    assert result.value.executor == "executor"
    assert "Traceback" in result.value.traceback
    assert "File" in result.value.traceback

    assert "Traceback" in str(result)


@pytest.mark.unit
def test_should_construct_critical_error_from_base_exception_and_valid_input_parameters():
    valid_input_parameters = ("hola", 2)

    error = CriticalError(
        Exception("my_trace"), input_parameters=valid_input_parameters
    )

    assert error.message == "Exception: my_trace"
    assert "Input Parameters: {'param_1': 'hola', 'param_2': 2}" in error.__repr__()


@pytest.mark.unit
def test_should_construct_critical_error_from_base_exception_and_valid_input_parameters_filtering_parameters():
    valid_input_parameters = ("hola", 2)

    error = CriticalError(
        Exception("my_trace"),
        input_parameters=valid_input_parameters,
        filter_parameters=["param_2"],
    )

    assert error.message == "Exception: my_trace"
    assert "Input Parameters: {'param_1': 'hola'}" in error.__repr__()



import atexit
from datetime import datetime, timedelta

from apscheduler.schedulers import (
    SchedulerAlreadyRunningError,
    SchedulerNotRunningError,
)
from apscheduler.schedulers.blocking import BlockingScheduler

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger

from petisco.tasks.config.config_tasks import ConfigTasks
from petisco.tasks.domain.interface_task_executor import TaskExecutor


class APSchedulerTaskExecutor(TaskExecutor):
    def __init__(self, scheduler: BlockingScheduler = BackgroundScheduler()):
        self.scheduler = scheduler

    def start(self, config_tasks: ConfigTasks):
        for task in config_tasks.tasks.values():
            self._config_task(task)
        try:
            self.scheduler.start()
            atexit.register(lambda: self.stop())
        except SchedulerAlreadyRunningError:
            pass

    def _config_task(self, task):
        if task.type == "recurring":
            self._config_recurring_task(task)
        elif task.type == "scheduled":
            self._config_scheduled_task(task)
        elif task.type == "cron":
            self._config_cron_task(task)
        else:
            self._config_instant_task(task)

    def _config_recurring_task(self, task):
        now = datetime.utcnow()
        start_date = now + timedelta(0, task.run_in)
        self.scheduler.add_job(
            func=task.get_handler(),
            start_date=start_date,
            trigger="interval",
            seconds=task.interval,
        )

    def _config_scheduled_task(self, task):
        now = datetime.utcnow()
        start_date = now + timedelta(0, task.run_in)
        self.scheduler.add_job(
            func=task.get_handler(), trigger="date", run_date=start_date
        )

    def _config_cron_task(self, task):
        self.scheduler.add_job(
            func=task.get_handler(), trigger=CronTrigger.from_crontab(task.every)
        )

    def _config_instant_task(self, task):
        self.scheduler.add_job(func=task.get_handler())

    def stop(self):
        try:
            self.scheduler.shutdown()
        except SchedulerNotRunningError:
            pass



from meiga import Success, Failure

from petisco import WebhookResultId, Response, TimeoutRequestError
from petisco.webhooks.webhook.domain.webhook_result import WebhookResult
from tests.modules.webook.mothers.webhook_mother import WebhookMother


class WebhookResultMother:
    @staticmethod
    def success():
        return WebhookResult.from_dict(
            {
                "webhook_result_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "webhook_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "sent_on": "2020-11-18 11:38:37.156162",
                "response": {
                    "headers": None,
                    "body": "",
                    "status_code": 200,
                    "completed_in_ms": 10,
                },
                "request": {
                    "headers": '{"X-Petisco-Event": "user_created", "X-Petisco-Event-Version": "1", "X-Petisco-Delivery": "a026531a-8512-4599-9491-81909ea8e31b", "X-Petisco-WebhookId": "1baa14dc-bda9-400f-b45b-428f26532f1d", "X-Petisco-Request-Timestamp": "2020-11-18 11:38:37.156162", "User-Agent": "Petisco-Hookshoot/", "apikey": "\'b0b905d6-228f-44bf-a130-c85d7aecd765"}',
                    "body": '{"payload": "ok"}',
                },
                "is_success": True,
            }
        )

    @staticmethod
    def success_from_create():
        webhook_result = WebhookResultMother.success()
        return WebhookResult.create(
            webhook=WebhookMother.default(),
            webhook_result_id=WebhookResultId.generate(),
            sent_on=webhook_result.sent_on,
            request_headers={
                "X-Petisco-Event": "user_created",
                "X-Petisco-Event-Version": "1",
                "X-Petisco-Delivery": "a026531a-8512-4599-9491-81909ea8e31b",
                "X-Petisco-WebhookId": "1baa14dc-bda9-400f-b45b-428f26532f1d",
                "X-Petisco-Request-Timestamp": "2020-11-18 11:38:37.156162",
                "User-Agent": "Petisco-Hookshoot/",
                "apikey": "'b0b905d6-228f-44bf-a130-c85d7aecd765",
            },
            request_body={"payload": "ok"},
            result=Success(Response(200)),
        )

    @staticmethod
    def failure_from_create():
        webhook_result = WebhookResultMother.success()
        return WebhookResult.create(
            webhook=WebhookMother.default(),
            webhook_result_id=WebhookResultId.generate(),
            sent_on=webhook_result.sent_on,
            request_headers={
                "X-Petisco-Event": "user_created",
                "X-Petisco-Event-Version": "1",
                "X-Petisco-Delivery": "a026531a-8512-4599-9491-81909ea8e31b",
                "X-Petisco-WebhookId": "1baa14dc-bda9-400f-b45b-428f26532f1d",
                "X-Petisco-Request-Timestamp": "2020-11-18 11:38:37.156162",
                "User-Agent": "Petisco-Hookshoot/",
                "apikey": "'b0b905d6-228f-44bf-a130-c85d7aecd765",
            },
            request_body={"payload": "ok"},
            result=Failure(TimeoutRequestError()),
        )

    @staticmethod
    def failure():
        return WebhookResult.from_dict(
            {
                "webhook_result_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "webhook_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "sent_on": "2020-11-18 11:41:18.140811",
                "response": {"headers": None, "body": None, "status_code": 503},
                "request": {
                    "headers": '{"X-Petisco-Event": "user_created", "X-Petisco-Event-Version": "1", "X-Petisco-Delivery": "14ac35e8-0483-40b2-b308-104309b7407b", "X-Petisco-WebhookId": "8d5dc972-a6c8-4c87-8f81-a4cb61dcb6a3", "X-Petisco-Request-Timestamp": "2020-11-18 11:41:18.140811", "User-Agent": "Petisco-Hookshoot/", "apikey": "\'b0b905d6-228f-44bf-a130-c85d7aecd765"}',
                    "body": '{"payload": "ok"}',
                },
                "is_success": False,
            }
        )

    @staticmethod
    def without_response():
        return WebhookResult.from_dict(
            {
                "webhook_result_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "webhook_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "sent_on": "2020-11-18 11:38:37.156162",
                "response": None,
                "request": {
                    "headers": '{"X-Petisco-Event": "user_created", "X-Petisco-Event-Version": "1", "X-Petisco-Delivery": "a026531a-8512-4599-9491-81909ea8e31b", "X-Petisco-WebhookId": "1baa14dc-bda9-400f-b45b-428f26532f1d", "X-Petisco-Request-Timestamp": "2020-11-18 11:38:37.156162", "User-Agent": "Petisco-Hookshoot/", "apikey": "\'b0b905d6-228f-44bf-a130-c85d7aecd765"}',
                    "body": '{"payload": "ok"}',
                },
                "is_success": True,
            }
        )

    @staticmethod
    def without_request():
        return WebhookResult.from_dict(
            {
                "webhook_result_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "webhook_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "sent_on": "2020-11-18 11:38:37.156162",
                "response": {
                    "headers": None,
                    "body": "",
                    "status_code": 200,
                    "completed_in_ms": 10,
                },
                "request": None,
                "is_success": True,
            }
        )

    @staticmethod
    def without_request_headers():
        return WebhookResult.from_dict(
            {
                "webhook_result_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "webhook_id": "a026531a-8512-4599-9491-81909ea8e31b",
                "sent_on": "2020-11-18 11:38:37.156162",
                "response": {
                    "headers": None,
                    "body": "",
                    "status_code": 200,
                    "completed_in_ms": 10,
                },
                "request": {"headers": None, "body": '{"payload": "ok"}'},
                "is_success": True,
            }
        )



import pytest

from petisco.notifier.infrastructure.slack.slack_notifier_message_converter import (
    SlackNotifierMessageConverter,
)


@pytest.mark.integration
def test_should_convert_a_basic_notifier_message(
    given_any_basic_notifier_message, given_any_message
):
    converter = SlackNotifierMessageConverter()
    blocks = converter.convert(given_any_basic_notifier_message)
    assert len(blocks) == 1
    assert "text" in blocks[0].keys()
    assert given_any_message in blocks[0]["text"]["text"]


@pytest.mark.integration
def test_should_convert_a_notifier_message_with_info_id(
    given_any_notifier_message_with_info_id, given_any_message
):
    converter = SlackNotifierMessageConverter()
    blocks = converter.convert(given_any_notifier_message_with_info_id)
    assert len(blocks) == 3
    assert "information" in blocks[0]["text"]["text"]
    assert "Client" in blocks[0]["text"]["text"]
    assert "User" in blocks[0]["text"]["text"]
    assert "divider" in blocks[1].values()
    assert "text" in blocks[2].keys()
    assert given_any_message in blocks[2]["text"]["text"]


@pytest.mark.integration
def test_should_convert_a_complete_notifier_message(
    given_any_complete_notifier_message, given_any_message, given_any_title
):
    converter = SlackNotifierMessageConverter()
    blocks = converter.convert(given_any_complete_notifier_message)
    assert len(blocks) == 5
    assert "Petisco" in blocks[0]["text"]["text"]
    assert "Application" in blocks[0]["text"]["text"]
    assert "information" in blocks[1]["text"]["text"]
    assert "Client" in blocks[1]["text"]["text"]
    assert "User" in blocks[1]["text"]["text"]
    assert "divider" in blocks[2].values()
    assert "text" in blocks[3].keys()
    assert given_any_title in blocks[3]["text"]["text"]
    assert "text" in blocks[4].keys()
    assert given_any_message in blocks[4]["text"]["text"]


@pytest.mark.integration
def test_should_convert_a_notifier_exception_message(
    given_any_complete_notifier_exception_message,
    given_any_exception,
    given_any_traceback,
):
    converter = SlackNotifierMessageConverter()
    blocks = converter.convert(given_any_complete_notifier_exception_message)
    assert len(blocks) == 7
    assert "Petisco" in blocks[0]["text"]["text"]
    assert "Application" in blocks[0]["text"]["text"]
    assert "information" in blocks[1]["text"]["text"]
    assert "Client" in blocks[1]["text"]["text"]
    assert "User" in blocks[1]["text"]["text"]
    assert "divider" in blocks[2].values()
    assert "text" in blocks[3].keys()
    assert str(given_any_exception) in blocks[3]["text"]["text"]
    assert "text" in blocks[4].keys()
    assert "Executor" in blocks[4]["text"]["text"]
    assert "text" in blocks[5].keys()
    assert "text" in blocks[6].keys()
    assert str(given_any_traceback) in blocks[6]["text"]["text"]


@pytest.mark.integration
def test_should_truncate_a_long_block_when_convert_slack_notifier(
    given_long_traceback_complete_notifier_exception_message,
):
    converter = SlackNotifierMessageConverter()
    blocks = converter.convert(given_long_traceback_complete_notifier_exception_message)
    assert "text" in blocks[5].keys()
    assert len(blocks[5]["text"]["text"]) < 3000



from time import sleep
from typing import Callable

import pytest
from meiga import isSuccess

from petisco import (
    Event,
    subscriber_handler,
    RabbitMQEventPublisher,
    RabbitMQEventSubscriber,
    ConfigEventSubscriber,
    RabbitMqConnector,
    RoutingKey,
    ERROR,
    WARNING,
)

from petisco.event.shared.infrastructure.rabbitmq.rabbitmq_is_running_locally import (
    rabbitmq_is_running_locally,
)
from tests.fixtures import TrackedEventsSpy
from tests.modules.unit.mocks.fake_logger import FakeLogger
from tests.modules.unit.mocks.log_message_mother import LogMessageMother


def await_for_subscriber():
    sleep(2.0)


def await_for_events():
    sleep(1.5)


@pytest.fixture
def given_any_publisher_and_subscriber(
    given_any_publisher,
    given_any_subscriber,
    given_random_organization,
    given_random_service,
    given_random_topic,
):
    def _given_any_publisher_and_subscriber(
        subscriber_handler: Callable,
        organization: str = None,
        service: str = None,
        topic: str = None,
    ):

        organization = organization if organization else given_random_organization
        service = service if service else given_random_service
        topic = topic if topic else given_random_topic

        publisher = given_any_publisher(organization, service, topic)
        subscriber = given_any_subscriber(
            organization, service, topic, subscriber_handler
        )
        return publisher, subscriber

    return _given_any_publisher_and_subscriber


@pytest.fixture
def given_any_publisher():
    def _given_any_publisher(
        given_random_organization, given_random_service, given_random_topic
    ):
        publisher = RabbitMQEventPublisher(
            connector=RabbitMqConnector(),
            organization=given_random_organization,
            service=given_random_service,
            topic=given_random_topic,
        )
        return publisher

    return _given_any_publisher


@pytest.fixture
def given_any_subscriber():
    def _given_any_subscriber(
        given_random_organization,
        given_random_service,
        given_random_topic,
        subscriber_handler: Callable,
    ):
        subscriber = RabbitMQEventSubscriber(
            connector=RabbitMqConnector(),
            subscribers={
                "petisco": ConfigEventSubscriber(
                    organization=given_random_organization,
                    service=given_random_service,
                    topic=given_random_topic,
                    handler=subscriber_handler,
                )
            },
        )
        return subscriber

    return _given_any_subscriber


@pytest.mark.integration
@pytest.mark.skipif(
    not rabbitmq_is_running_locally(), reason="RabbitMQ is not running locally"
)
def test_should_subscriber_handler_receive_one_event(
    given_any_petisco, make_user_created_event, given_any_publisher_and_subscriber
):
    event = make_user_created_event()
    tracked_events_spy = TrackedEventsSpy()

    @subscriber_handler()
    def main_handler(event: Event):
        tracked_events_spy.append(event)
        return isSuccess

    publisher, subscriber = given_any_publisher_and_subscriber(main_handler)

    subscriber.start()

    await_for_subscriber()

    publisher.publish(event)

    await_for_events()

    tracked_events_spy.assert_number_events(1)

    subscriber.stop()


@pytest.mark.integration
@pytest.mark.skipif(
    not rabbitmq_is_running_locally(), reason="RabbitMQ is not running locally"
)
def test_should_subscriber_handler_always_simulate_a_nack(
    given_any_petisco, make_user_created_event, given_any_publisher_and_subscriber
):
    logger = FakeLogger()
    event = make_user_created_event()
    tracked_events_spy = TrackedEventsSpy()

    @subscriber_handler(logger=logger, percentage_simulate_nack=1.0)
    def main_handler(event: Event):
        tracked_events_spy.append(event)
        return isSuccess

    publisher, subscriber = given_any_publisher_and_subscriber(main_handler)

    subscriber.start()

    await_for_subscriber()

    publisher.publish(event)

    await_for_events()

    tracked_events_spy.assert_number_events(0)

    second_logging_message = logger.get_logging_messages()[1]

    assert second_logging_message == (
        WARNING,
        LogMessageMother.get_subscriber(
            operation="main_handler",
            message="Message rejected (Simulation rejecting 100.0% of the messages)",
        ).to_dict(),
    )

    subscriber.stop()


@pytest.mark.integration
@pytest.mark.skipif(
    not rabbitmq_is_running_locally(), reason="RabbitMQ is not running locally"
)
def test_should_subscriber_handler_always_returns_nack_filtering_by_invalid_routing_key(
    given_any_petisco, make_user_created_event, given_any_publisher_and_subscriber
):
    logger = FakeLogger()
    event = make_user_created_event()
    tracked_events_spy = TrackedEventsSpy()
    invalid_routing_key = "invalid.routing.key"

    @subscriber_handler(logger=logger, filter_routing_key=invalid_routing_key)
    def main_handler(event: Event):
        tracked_events_spy.append(event)
        return isSuccess

    publisher, subscriber = given_any_publisher_and_subscriber(main_handler)

    subscriber.start()

    await_for_subscriber()

    publisher.publish(event)

    await_for_events()

    tracked_events_spy.assert_number_events(0)

    second_logging_message = logger.get_logging_messages()[1]

    assert second_logging_message == (
        WARNING,
        LogMessageMother.get_subscriber(
            operation="main_handler",
            message=f"Message rejected (filtering by routing_key {invalid_routing_key})",
        ).to_dict(),
    )
    subscriber.stop()


@pytest.mark.integration
@pytest.mark.skipif(
    not rabbitmq_is_running_locally(), reason="RabbitMQ is not running locally"
)
def test_should_subscriber_handler_receive_one_event_and_obtain_the_routing_key(
    given_any_petisco, make_user_created_event, given_any_publisher_and_subscriber
):
    event = make_user_created_event()
    tracked_events_spy = TrackedEventsSpy()
    expected_organization = "acmeorganization"

    @subscriber_handler()
    def main_handler(event: Event, routing_key: RoutingKey):
        assert routing_key.match(organization=expected_organization)
        tracked_events_spy.append(event)
        return isSuccess

    publisher, subscriber = given_any_publisher_and_subscriber(
        main_handler, organization=expected_organization
    )

    subscriber.start()

    await_for_subscriber()

    publisher.publish(event)

    await_for_events()

    tracked_events_spy.assert_number_events(1)

    subscriber.stop()


@pytest.mark.integration
@pytest.mark.skipif(
    not rabbitmq_is_running_locally(), reason="RabbitMQ is not running locally"
)
def test_should_subscriber_handler_return_a_failure_with_unknown_error_when_raise_an_uncontrolled_exception(
    given_any_petisco, make_user_created_event, given_any_publisher_and_subscriber
):
    event = make_user_created_event()
    expected_organization = "acmeorganization"

    logger = FakeLogger()

    @subscriber_handler(logger=logger)
    def main_handler(event: Event, routing_key: RoutingKey):
        raise RuntimeError("uncontrolled exception")

    publisher, subscriber = given_any_publisher_and_subscriber(
        main_handler, organization=expected_organization
    )

    subscriber.start()

    await_for_subscriber()

    publisher.publish(event)

    await_for_events()

    subscriber.stop()

    second_logging_message = logger.get_logging_messages()[1]
    assert second_logging_message[0] == ERROR
    assert (
        "Result[status: failure | value: UnknownError (main_handler (Subscriber)): RuntimeError: uncontrolled exception."
        in second_logging_message[1]["data"]["message"]
    )
    assert "Input Parameters:" in second_logging_message[1]["data"]["message"]



import pytest

from petisco.domain.aggregate_roots.info_id import InfoId
from petisco.domain.value_objects.correlation_id import CorrelationId


@pytest.mark.unit
def test_should_check_info_id_to_dict_method(
    given_any_info_id, given_any_client_id, given_any_user_id, given_any_correlation_id
):
    info_id = given_any_info_id

    assert info_id.to_dict() == {
        "client_id": given_any_client_id.value,
        "user_id": given_any_user_id.value,
        "correlation_id": given_any_correlation_id.value,
        "ip": None,
    }


@pytest.mark.unit
def test_should_check_info_id_get_http_headers_method(given_any_info_id):
    info_id = given_any_info_id

    headers = info_id.get_http_headers()

    assert headers["X-Onboarding-Clientid"] == info_id.client_id.value
    assert headers["X-Onboarding-Userid"] == info_id.user_id.value
    assert headers["X-Correlation-Id"] == info_id.correlation_id.value


@pytest.mark.unit
def test_should_create_info_id_from_headers(
    given_any_client_id, given_any_user_id, given_any_correlation_id
):
    headers = {
        "X-Onboarding-Clientid": given_any_client_id.value,
        "X-Onboarding-Userid": given_any_user_id.value,
        "X-Correlation-Id": given_any_correlation_id.value,
    }

    info_id = InfoId.from_headers(headers)

    assert headers["X-Onboarding-Clientid"] == info_id.client_id.value
    assert headers["X-Onboarding-Userid"] == info_id.user_id.value
    assert headers["X-Correlation-Id"] == info_id.correlation_id.value


@pytest.mark.unit
def test_should_create_info_id_from_headers_empty():
    headers = {}

    info_id = InfoId.from_headers(headers)

    assert info_id.client_id is None
    assert info_id.user_id is None
    assert info_id.correlation_id is not None  # autogenerated


@pytest.mark.unit
def test_should_create_info_id_from_headers_none():

    info_id = InfoId.from_headers(None)

    assert info_id.client_id is None
    assert info_id.user_id is None
    assert info_id.correlation_id is not None  # autogenerated


@pytest.mark.unit
def test_should_create_info_id_from_headers_with_only_correlation_id(
    given_any_correlation_id
):

    headers = {"X-Correlation-Id": given_any_correlation_id.value}

    info_id = InfoId.from_headers(headers)

    assert info_id.client_id is None
    assert info_id.user_id is None
    assert info_id.correlation_id == given_any_correlation_id


@pytest.mark.unit
def test_should_update_info_id_from_headers(given_any_info_id):

    external_correlation_id = CorrelationId.generate().value

    headers = {"X-Correlation-Id": external_correlation_id}

    info_id = given_any_info_id
    autogenerated_correlation_id = info_id.correlation_id

    returned_info_id = info_id.update_from_headers(headers)

    assert info_id.correlation_id.value == external_correlation_id
    assert info_id.correlation_id != autogenerated_correlation_id
    assert returned_info_id == info_id



from unittest import mock
from unittest.mock import PropertyMock

import pytest
from pika import BlockingConnection

from petisco import RabbitMqConnector

from petisco.fixtures import testing_with_rabbitmq


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_get_an_open_connection():

    connector = RabbitMqConnector()

    connection = connector.get_connection("test")

    assert connection.is_open

    connection.close()


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_throw_a_connection_error_exception_when_input_envvars_are_not_valid():

    connector = RabbitMqConnector()
    original_host = connector.host
    connector.host = "invalid"

    with pytest.raises(
        ConnectionError, match="RabbitMQConnector: Impossible to connect to host*"
    ):
        connector.get_connection("test")

    connector.host = original_host


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_throw_a_connection_error_exception_when_retries_and_is_not_possible_to_obtain_an_open_connection():

    connector = RabbitMqConnector()
    original_wait_seconds_retry = connector.wait_seconds_retry
    connector.wait_seconds_retry = 0.1

    connection = connector.get_connection("test")
    connection.close()

    with mock.patch.object(
        BlockingConnection, "is_open", new_callable=PropertyMock
    ) as mocker_is_open:
        mocker_is_open.return_value = False
        with pytest.raises(
            ConnectionError,
            match="RabbitMQConnector: Impossible to obtain a open connection with host*",
        ):
            connector.get_connection("test")

    connector.wait_seconds_retry = original_wait_seconds_retry


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_recover_from_connection_closed():

    connector = RabbitMqConnector()

    connection = connector.get_connection("test")
    connection.close()

    connection = connector.get_connection("test")

    assert connection.is_open

    connection.close()



import importlib
import sys
from typing import Optional, Dict, Any, Callable, List

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


@dataclass
class SingleConfigPersistence:
    type: str
    config: Optional[Callable] = None
    models: Optional[Dict[str, Any]] = None


@dataclass
class ConfigPersistence:

    configs: Optional[Dict[str, SingleConfigPersistence]] = None

    @staticmethod
    def from_dict(kdict):
        if not kdict or not isinstance(kdict, List):
            return ConfigPersistence()
        configs = {}
        for entry in kdict:
            config = (
                get_function_from_string(entry.get("config"))
                .handle(
                    on_failure=raise_petisco_config_exception,
                    failure_args=(entry, "persistence:config"),
                )
                .unwrap()
            )
            configs[entry["name"]] = SingleConfigPersistence(
                type=entry["type"], config=config, models=entry.get("models")
            )
        return ConfigPersistence(configs=configs)

    def get_models(self, persistence_entry):
        loaded_models = {}
        for name, model_string in self.configs[persistence_entry].models.items():
            mod_name, model_name = model_string.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            loaded_models[name] = getattr(mod, model_name)
        return loaded_models

    def get_import_database_models_func(self, persistence_entry):
        def _is_class_in_sqlalchemy_tables(class_model_name: str):
            # We need this to filter deletion in case of Model dependencies.
            # For instance, ClientConfigModel imports ClientModel on class model definition
            # It throws the following error:
            #  sqlalchemy.exc.InvalidRequestError: Table 'ClientConfig' is already defined for this MetaData instance.
            #  Specify 'extend_existing=True' to redefine options and columns on an existing Table object.
            # As a workaround, we only delete the module if it not exist in SQLAlchemy tables.
            from petisco.persistence.sqlalchemy.sqlalchemy_persistence import (
                SqlAlchemyPersistence,
            )

            is_in_tables = False
            for source in SqlAlchemyPersistence.get_instance().sources.values():
                Base = source.get("base")
                if not Base:
                    continue
                is_in_tables = class_model_name.replace("Model", "") in list(
                    Base.metadata.tables.keys()
                )
            return is_in_tables

        def _delete_module_if_already_imported(module_name: str, class_model_name: str):
            # Important. SqlAlchemy needs models to be imported
            # Base = SqlAlchemyPersistence.get_instance().sources["petisco"]["base"]
            # If module is already imported and Base has not tables, Tables won't be imported.
            # This, usually happens when start the application (e.g end2end tests)
            # e.g len(Base.metadata.tables) == 0
            # With this function we ensure that model can be imported to help SqlAlchemy loading models if is not imported yet
            if module_name in sys.modules and not _is_class_in_sqlalchemy_tables(
                class_model_name
            ):
                del sys.modules[module_name]

        def _import_database_models_func():
            for name, model_string in self.configs[persistence_entry].models.items():
                module_name, class_model_name = model_string.rsplit(".", 1)
                _delete_module_if_already_imported(module_name, class_model_name)
                importlib.import_module(module_name, class_model_name)

        return _import_database_models_func



import inspect
import traceback

from functools import wraps
from typing import Callable, Tuple, Dict, List, Any
from meiga import Result, Error, Success
from meiga.decorators import meiga

from petisco.event.bus.domain.interface_event_bus import IEventBus
from petisco.application.petisco import Petisco
from petisco.controller.errors.internal_http_error import InternalHttpError
from petisco.controller.errors.known_result_failure_handler import (
    KnownResultFailureHandler,
)

from petisco.domain.aggregate_roots.info_id import InfoId
from petisco.domain.errors.critical_error import CriticalError

from petisco.event.shared.domain.request_responded import RequestResponded
from petisco.frameworks.flask.flask_headers_provider import flask_headers_provider
from petisco.logger.interface_logger import ERROR, DEBUG
from petisco.controller.errors.http_error import HttpError
from petisco.notifier.domain.interface_notifier import INotifier
from petisco.notifier.domain.notifier_exception_message import NotifierExceptionMessage
from petisco.security.token_manager.not_implemented_token_manager import (
    NotImplementedTokenManager,
)
from petisco.security.token_manager.token_manager import TokenManager
from petisco.logger.log_message import LogMessage
from petisco.tools.timer import timer

DEFAULT_SUCCESS_MESSAGE = {"message": "OK"}, 200
DEFAULT_ERROR_MESSAGE = HttpError().handle()
DEFAULT_APP_NAME = "app-undefined"
DEFAULT_APP_VERSION = "version-undefined"
DEFAULT_LOGGER = None
DEFAULT_EVENT_BUS = None
DEFAULT_NOTIFIER = None


class _ControllerHandler:
    def __init__(
        self,
        app_name: str = DEFAULT_ERROR_MESSAGE,
        app_version: str = DEFAULT_APP_VERSION,
        logger=DEFAULT_LOGGER,
        notifier: INotifier = DEFAULT_NOTIFIER,
        token_manager: TokenManager = NotImplementedTokenManager(),
        success_handler: Callable[[Result], Tuple[Dict, int]] = None,
        error_handler: Callable[[Result], HttpError] = None,
        headers_provider: Callable = flask_headers_provider,
        logging_types_blacklist: List[Any] = [bytes],
        event_bus: IEventBus = DEFAULT_EVENT_BUS,
        send_request_responded_event: bool = False,
    ):
        """
        Parameters
        ----------
        app_name
            Application Name. If not specified it will get it from Petisco.get_app_version().
        app_version
            Application Version. If not specified it will get it from Petisco.get_app_version().
        logger
            A ILogger implementation. If not specified it will get it from Petisco.get_logger(). You can also use NotImplementedLogger
        notifier
            A INotifier implementation. If not specified it will get it from Petisco.get_notifier(). You can also use NotImplementedNotifier
        token_manager
            TokenManager object. Here, you can define how to deal with JWT Tokens
        success_handler
            Handler to deal with Success Results
        error_handler
            Handler to deal with Failure Results
        headers_provider
            Injectable function to provide headers. By default is used headers_provider
        logging_types_blacklist
            Logging Blacklist. Object of defined Type will not be logged. By default ( [bytes] ) bytes object won't be logged.
        event_bus
            A IEventBus implementation. If not specified it will get it from Petisco.get_event_bus().
        send_request_responded_event
            Boolean to select if RequestResponded event is send. It will use provided publisher
        """
        self.app_name = app_name
        self.app_version = app_version
        self.logger = logger
        self.notifier = notifier
        self.event_bus = event_bus
        self.token_manager = token_manager
        self.success_handler = success_handler
        self.error_handler = error_handler
        self.headers_provider = headers_provider
        self.logging_types_blacklist = logging_types_blacklist
        self.send_request_responded_event = send_request_responded_event

    def _check_app_name(self):
        if self.app_name == DEFAULT_ERROR_MESSAGE:
            self.app_name = Petisco.get_app_name()

    def _check_app_version(self):
        if self.app_version == DEFAULT_APP_VERSION:
            self.app_version = Petisco.get_app_version()

    def _check_logger(self):
        if self.logger == DEFAULT_LOGGER:
            self.logger = Petisco.get_logger()

    def _check_notifier(self):
        if self.notifier == DEFAULT_NOTIFIER:
            self.notifier = Petisco.get_notifier()

    def _check_publisher(self):
        if self.event_bus == DEFAULT_EVENT_BUS:
            self.event_bus = Petisco.get_event_bus()

    def _configure_petisco_dependencies(self):
        self.petisco = Petisco.get_instance()
        self._check_app_name()
        self._check_app_version()
        self._check_logger()
        self._check_notifier()
        self._check_publisher()

    def _get_success_message(self, result: Result):
        message = ""
        if isinstance(result.value, tuple(self.logging_types_blacklist)):
            message = f"Success result of type: {type(result.value).__name__}"
        else:
            message = f"{result}"
        return message

    def __call__(self, func, *args, **kwargs):
        @wraps(func)
        def wrapper(*args, **kwargs):
            @timer
            @meiga
            def run_controller(*args, **kwargs) -> Tuple[Result, float]:
                params = inspect.getfullargspec(func).args
                kwargs = {k: v for k, v in kwargs.items() if k in params}
                return func(*args, **kwargs)

            self._configure_petisco_dependencies()

            info_id = None
            is_success = False
            elapsed_time = None
            try:
                kwargs = add_petisco(kwargs)
                kwargs = add_headers(kwargs)
                result_kwargs = add_info_id(kwargs)

                self.notify(result_kwargs, info_id)

                log_message = LogMessage(
                    layer="controller", operation=f"{func.__name__}"
                )
                if result_kwargs.is_failure:
                    http_response = self.handle_failure(log_message, result_kwargs)
                else:
                    kwargs, info_id = result_kwargs.value
                    log_message.info_id = info_id
                    self.logger.log(
                        DEBUG, log_message.set_message("Processing Request")
                    )

                    result_controller, elapsed_time = run_controller(*args, **kwargs)

                    self.notify(result_controller, info_id)

                    if result_controller.is_failure:
                        http_response = self.handle_failure(
                            log_message, result_controller
                        )
                    else:
                        message = self._get_success_message(result_controller)
                        self.logger.log(DEBUG, log_message.set_message(message))
                        http_response = (
                            self.success_handler(result_controller)
                            if self.success_handler
                            else DEFAULT_SUCCESS_MESSAGE
                        )
                        is_success = True
            except Exception as e:
                log_message = LogMessage(
                    layer="controller", operation=f"{func.__name__}"
                )
                log_message.info_id = info_id if info_id else None
                message = f"Error {func.__name__}: {repr(e.__class__)} {e} | {traceback.format_exc()}"
                self.logger.log(ERROR, log_message.set_message(message))
                http_response = InternalHttpError().handle()
                self.notifier.publish(
                    NotifierExceptionMessage(
                        exception=e,
                        executor=func.__name__,
                        traceback=traceback.format_exc(),
                        info_id=info_id,
                        info_petisco=Petisco.get_info(),
                    )
                )

            if self.send_request_responded_event:

                request_responded = RequestResponded(
                    app_name=self.app_name,
                    app_version=self.app_version,
                    controller=f"{func.__name__}",
                    is_success=is_success,
                    http_response=http_response,
                    elapsed_time=elapsed_time,
                ).add_info_id(info_id)

                self.event_bus.publish(request_responded)

            return http_response

        def add_headers(kwargs):
            headers = self.headers_provider()
            kwargs = dict(kwargs, headers=headers)
            return kwargs

        def add_petisco(kwargs):
            kwargs = dict(kwargs, petisco=self.petisco)
            return kwargs

        @meiga
        def add_info_id(kwargs) -> Result[Tuple[str, InfoId], Error]:
            headers = kwargs.get("headers", {})

            info_id = self.token_manager.execute(headers).unwrap_or_return()

            kwargs = dict(kwargs, info_id=info_id)
            return Success((kwargs, info_id))

        return wrapper

    @meiga
    def notify(self, result, info_id: InfoId = None):
        if result.is_failure:
            error = result.value
            if issubclass(error.__class__, CriticalError):
                self.notifier.publish(
                    NotifierExceptionMessage(
                        exception=error.exception,
                        executor=error.executor,
                        input_parameters=error.input_parameters,
                        traceback=error.traceback,
                        info_id=info_id,
                        info_petisco=Petisco.get_info(),
                    )
                )

    def handle_failure(
        self, log_message: LogMessage, result: Result
    ) -> Tuple[dict, int]:
        self.logger.log(ERROR, log_message.set_message(f"{result}"))
        known_result_failure_handler = KnownResultFailureHandler(result)
        http_response = DEFAULT_ERROR_MESSAGE
        if not known_result_failure_handler.is_a_result_known_error:
            if self.error_handler:
                http_error = self.error_handler(result)

                if not issubclass(http_error.__class__, HttpError):
                    raise TypeError(
                        "Returned object from error_handler must be subclasses of HttpError"
                    )

                http_response = http_error.handle()
        else:
            http_response = known_result_failure_handler.http_error.handle()
        return http_response


controller_handler = _ControllerHandler



import json

from meiga import Result


class WebhookResponseResult:
    @staticmethod
    def from_result(result: Result):
        if result.is_success:
            response = result.unwrap()
            return WebhookResponseResult(
                headers=response.headers,
                body=response.content,
                status_code=response.status_code,
                completed_in_ms=response.completed_in_ms,
            )

        error = result.value
        return WebhookResponseResult(
            headers=error.headers,
            body=error.content,
            status_code=error.status_code,
            completed_in_ms=error.completed_in_ms,
        )

    @staticmethod
    def from_dict(kdict: dict):
        return WebhookResponseResult(
            headers=json.loads(kdict.get("headers")) if kdict.get("headers") else None,
            body=json.loads(kdict.get("body")) if kdict.get("body") else None,
            status_code=kdict.get("status_code"),
            completed_in_ms=kdict.get("completed_in_ms"),
        )

    def to_dict(self):
        return {
            "headers": json.dumps(self.headers) if self.headers else None,
            "body": json.dumps(self.body) if self.body else None,
            "status_code": self.status_code,
            "completed_in_ms": self.completed_in_ms,
        }

    def __init__(
        self, headers: dict, body: dict, status_code: int, completed_in_ms: float = None
    ):
        self._set_headers(headers)
        self._set_body(body)
        self.status_code = status_code
        self.completed_in_ms = completed_in_ms
        super().__init__()

    def _set_headers(self, headers):
        try:
            self.headers = dict(headers)
        except Exception:
            self.headers = {}

    def _set_body(self, body):
        try:
            self.body = dict(body)
        except Exception:
            self.body = {}

    def __repr__(self):
        return json.dumps(self.to_dict())

    def __eq__(self, other):
        if issubclass(other.__class__, self.__class__) or issubclass(
            self.__class__, other.__class__
        ):
            return self.to_dict() == other.to_dict()
        else:
            return False



import pytest

from petisco import IService


@pytest.mark.unit
def test_should_inherit_from_service():
    class Service(IService):
        pass

    service = Service()

    assert service.info() == {"name": "Service"}


@pytest.mark.unit
def test_should_inherit_from_service_and_overwrite_info_method():
    class Service(IService):
        def info(self):
            return None

    service = Service()

    assert service.info() is None



import pytest

from petisco import IAppService


@pytest.mark.unit
def test_should_inherit_from_application_service():
    class AppService(IAppService):
        pass

    application_service = AppService()

    assert application_service.info() == {"name": "AppService"}


@pytest.mark.unit
def test_should_inherit_from_application_service_and_overwrite_info_method():
    class AppService(IAppService):
        def info(self):
            return None

    application_service = AppService()

    assert application_service.info() is None



from typing import Dict


class ConsumerDerivedAction:
    def __init__(
        self,
        action: str = None,
        exchange_name: str = None,
        routing_key: str = None,
        headers: Dict = None,
    ):
        self.action = action
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self.headers = headers

    def to_dict(self):
        if not self.action:
            return {}
        else:
            return {
                "action": self.action,
                "exchange_name": self.exchange_name,
                "routing_key": self.routing_key,
                "headers": self.headers,
            }



import copy
from time import sleep

import pytest
from meiga import isSuccess, isFailure, Result, Error

from petisco import Event
from petisco.event.shared.domain.event_subscriber import EventSubscriber

from petisco.fixtures.testing_decorators import testing_with_rabbitmq
from tests.modules.event.mothers.event_user_created_mother import EventUserCreatedMother
from tests.modules.event.mothers.rabbitmq_event_bus_mother import RabbitMqEventBusMother
from tests.modules.event.mothers.rabbitmq_event_configurer_mother import (
    RabbitMqEventConfigurerMother,
)
from tests.modules.event.mothers.rabbitmq_event_consumer_mother import (
    RabbitMqEventConsumerMother,
)
from tests.modules.event.spies.spy_events import SpyEvents


@pytest.mark.integration
@testing_with_rabbitmq
@pytest.mark.parametrize(
    "max_retries_allowed,expected_number_event_consumed,simulated_results",
    [
        (0, 1, [isFailure]),
        (1, 2, [isFailure, isSuccess]),
        (2, 3, [isFailure, isFailure, isSuccess]),
        (3, 4, [isFailure, isFailure, isFailure, isSuccess]),
        (4, 5, [isFailure, isFailure, isFailure, isFailure, isSuccess]),
    ],
)
def test_should_publish_consume_and_retry_event_from_rabbitmq_when_fail_consumer(
    max_retries_allowed, expected_number_event_consumed, simulated_results
):
    spy = SpyEvents()

    def assert_consumer(event: Event) -> Result[bool, Error]:
        spy.append(event)
        result = simulated_results.pop(0)
        return result

    event = EventUserCreatedMother.random()
    subscribers = [
        EventSubscriber(
            event_name=event.event_name,
            event_version=event.event_version,
            handlers=[assert_consumer],
        )
    ]

    configurer = RabbitMqEventConfigurerMother.with_retry_ttl_10ms()
    configurer.configure_subscribers(subscribers)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    consumer = RabbitMqEventConsumerMother.with_max_retries(max_retries_allowed)
    consumer.add_subscribers(subscribers)
    consumer.start()

    sleep(1.0)

    consumer.stop()
    configurer.clear()

    spy.assert_number_unique_events(1)
    spy.assert_first_event(event)
    spy.assert_last_event(event)
    spy.assert_count_by_event_id(event.event_id, expected_number_event_consumed)


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_publish_consume_and_retry_event_with_two_handlers_from_rabbitmq():

    spy_consumer_1 = SpyEvents()
    spy_consumer_2 = SpyEvents()

    def assert_consumer_1(event: Event) -> Result[bool, Error]:
        spy_consumer_1.append(event)
        return isSuccess

    def assert_consumer_2(event: Event) -> Result[bool, Error]:
        spy_consumer_2.append(event)
        return isSuccess

    event = EventUserCreatedMother.random()
    subscribers = [
        EventSubscriber(
            event_name=event.event_name,
            event_version=event.event_version,
            handlers=[assert_consumer_1, assert_consumer_2],
        )
    ]

    configurer = RabbitMqEventConfigurerMother.default()
    configurer.configure_subscribers(subscribers)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    consumer = RabbitMqEventConsumerMother.default()
    consumer.add_subscribers(subscribers)
    consumer.start()

    sleep(1.0)

    consumer.stop()
    configurer.clear()

    spy_consumer_1.assert_number_unique_events(1)
    spy_consumer_1.assert_first_event(event)
    spy_consumer_1.assert_count_by_event_id(event.event_id, 1)

    spy_consumer_2.assert_number_unique_events(1)
    spy_consumer_2.assert_first_event(event)
    spy_consumer_2.assert_count_by_event_id(event.event_id, 1)


@pytest.mark.integration
@testing_with_rabbitmq
@pytest.mark.parametrize(
    "max_retries_allowed,expected_number_event_consumed,simulated_results",
    [
        (0, 1, [isFailure]),
        (1, 2, [isFailure, isSuccess]),
        (2, 3, [isFailure, isFailure, isSuccess]),
        (3, 4, [isFailure, isFailure, isFailure, isSuccess]),
        (4, 5, [isFailure, isFailure, isFailure, isFailure, isSuccess]),
    ],
)
def test_should_publish_consume_and_retry_event_with_two_handlers_from_rabbitmq_when_fail_consumer(
    max_retries_allowed, expected_number_event_consumed, simulated_results
):

    spy_consumer_1 = SpyEvents()
    spy_consumer_2 = SpyEvents()

    simulated_results_1 = copy.deepcopy(simulated_results)
    simulated_results_2 = copy.deepcopy(simulated_results)

    def assert_consumer_1(event: Event) -> Result[bool, Error]:
        spy_consumer_1.append(event)
        result = simulated_results_1.pop(0)
        return result

    def assert_consumer_2(event: Event) -> Result[bool, Error]:
        spy_consumer_2.append(event)
        result = simulated_results_2.pop(0)
        return result

    event = EventUserCreatedMother.random()
    subscribers = [
        EventSubscriber(
            event_name=event.event_name,
            event_version=event.event_version,
            handlers=[assert_consumer_1, assert_consumer_2],
        )
    ]

    configurer = RabbitMqEventConfigurerMother.with_retry_ttl_10ms()
    configurer.configure_subscribers(subscribers)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    consumer = RabbitMqEventConsumerMother.with_max_retries(max_retries_allowed)
    consumer.add_subscribers(subscribers)
    consumer.start()

    sleep(1.5)

    consumer.stop()
    configurer.clear()

    print(f"num events: {len(spy_consumer_1.events)} - {spy_consumer_1}")
    print(f"num events: {len(spy_consumer_2.events)} - {spy_consumer_2}")

    spy_consumer_1.assert_number_unique_events(1)
    spy_consumer_1.assert_first_event(event)
    spy_consumer_1.assert_count_by_event_id(
        event.event_id, expected_number_event_consumed
    )

    spy_consumer_2.assert_number_unique_events(1)
    spy_consumer_2.assert_first_event(event)
    spy_consumer_2.assert_count_by_event_id(
        event.event_id, expected_number_event_consumed
    )


@pytest.mark.integration
@testing_with_rabbitmq
@pytest.mark.parametrize(
    "max_retries_allowed,expected_number_event_consumed,simulated_results",
    [
        (0, 1, [isFailure]),
        (1, 2, [isFailure, isSuccess]),
        (2, 3, [isFailure, isFailure, isSuccess]),
        (3, 4, [isFailure, isFailure, isFailure, isSuccess]),
        (4, 5, [isFailure, isFailure, isFailure, isFailure, isSuccess]),
    ],
)
def test_should_publish_consume_and_retry_event_not_affecting_store_queue_from_rabbitmq_when_fail_handler_consumer(
    max_retries_allowed, expected_number_event_consumed, simulated_results
):

    spy_consumer_event_store = SpyEvents()
    spy_consumer_handler = SpyEvents()

    def assert_consumer_event_store(event: Event) -> Result[bool, Error]:
        spy_consumer_event_store.append(event)
        return isSuccess

    def assert_consumer_handler(event: Event) -> Result[bool, Error]:
        spy_consumer_handler.append(event)
        result = simulated_results.pop(0)
        return result

    event = EventUserCreatedMother.random()
    subscribers = [
        EventSubscriber(
            event_name=event.event_name,
            event_version=event.event_version,
            handlers=[assert_consumer_handler],
        )
    ]

    configurer = RabbitMqEventConfigurerMother.with_retry_ttl_10ms()
    configurer.configure_subscribers(subscribers)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    consumer = RabbitMqEventConsumerMother.with_max_retries(max_retries_allowed)
    consumer.add_subscribers(subscribers)
    consumer.add_handler_on_queue("store", assert_consumer_event_store)

    consumer.start()

    sleep(1.0)

    consumer.stop()
    configurer.clear()

    spy_consumer_event_store.assert_number_unique_events(1)
    spy_consumer_event_store.assert_first_event(event)
    spy_consumer_event_store.assert_count_by_event_id(event.event_id, 1)

    spy_consumer_handler.assert_number_unique_events(1)
    spy_consumer_handler.assert_first_event(event)
    spy_consumer_handler.assert_count_by_event_id(
        event.event_id, expected_number_event_consumed
    )


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_publish_consume_retry_and_send_to_dead_letter_event_from_rabbitmq_when_fail_consumer():
    max_retries_allowed = 2
    expected_number_event_consumed = 3

    spy = SpyEvents()
    spy_dead_letter = SpyEvents()

    def assert_consumer(event: Event) -> Result[bool, Error]:
        spy.append(event)
        return isFailure

    event = EventUserCreatedMother.random()
    subscriber = EventSubscriber(
        event_name=event.event_name,
        event_version=event.event_version,
        handlers=[assert_consumer],
    )
    subscribers = [subscriber]

    configurer = RabbitMqEventConfigurerMother.with_retry_ttl_10ms()
    configurer.configure_subscribers(subscribers)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    consumer = RabbitMqEventConsumerMother.with_max_retries(max_retries_allowed)
    consumer.add_subscribers(subscribers)

    def dead_letter_consumer(event: Event) -> Result[bool, Error]:
        spy_dead_letter.append(event)
        return isSuccess

    consumer.add_subscriber_on_dead_letter(subscriber, dead_letter_consumer)

    consumer.start()

    sleep(1.5)

    consumer.stop()
    configurer.clear()

    spy.assert_number_unique_events(1)
    spy.assert_first_event(event)
    spy.assert_count_by_event_id(event.event_id, expected_number_event_consumed)

    spy_dead_letter.assert_number_unique_events(1)
    spy_dead_letter.assert_first_event(event)
    spy_dead_letter.assert_count_by_event_id(event.event_id, 1)


@pytest.mark.integration
@testing_with_rabbitmq
@pytest.mark.parametrize(
    "max_retries_allowed, expected_number_event_consumed_by_store, expected_number_event_consumed_by_handler_1, expected_number_event_consumed_by_handler_2,simulated_results_store, simulated_results_handler_1, simulated_results_handler_2",
    [
        (1, 2, 1, 1, [isFailure, isSuccess], [isSuccess], [isSuccess]),
        (1, 1, 2, 1, [isSuccess], [isFailure, isSuccess], [isSuccess]),
        (1, 1, 1, 2, [isSuccess], [isSuccess], [isFailure, isSuccess]),
        (1, 2, 2, 1, [isFailure, isSuccess], [isFailure, isSuccess], [isSuccess]),
        (2, 2, 1, 1, [isFailure, isSuccess], [isSuccess], [isSuccess]),
        (2, 1, 2, 1, [isSuccess], [isFailure, isSuccess], [isSuccess]),
        (2, 1, 1, 2, [isSuccess], [isSuccess], [isFailure, isSuccess]),
        (2, 2, 2, 1, [isFailure, isSuccess], [isFailure, isSuccess], [isSuccess]),
        (
            2,
            2,
            2,
            2,
            [isFailure, isSuccess],
            [isFailure, isSuccess],
            [isFailure, isSuccess],
        ),
        (3, 3, 1, 1, [isFailure, isFailure, isSuccess], [isSuccess], [isSuccess]),
        (3, 1, 3, 1, [isSuccess], [isFailure, isFailure, isSuccess], [isSuccess]),
        (3, 1, 1, 3, [isSuccess], [isSuccess], [isFailure, isFailure, isSuccess]),
    ],
)
def test_should_publish_consume_and_retry_event_not_affecting_other_queue_including_store_queue_from_rabbitmq(
    max_retries_allowed,
    expected_number_event_consumed_by_store,
    expected_number_event_consumed_by_handler_1,
    expected_number_event_consumed_by_handler_2,
    simulated_results_store,
    simulated_results_handler_1,
    simulated_results_handler_2,
):

    spy_consumer_event_store = SpyEvents()
    spy_consumer_handler_1 = SpyEvents()
    spy_consumer_handler_2 = SpyEvents()

    def assert_consumer_event_store(event: Event) -> Result[bool, Error]:
        spy_consumer_event_store.append(event)
        result = simulated_results_store.pop(0)
        return result

    def assert_consumer_handler_1(event: Event) -> Result[bool, Error]:
        spy_consumer_handler_1.append(event)
        result = simulated_results_handler_1.pop(0)
        return result

    def assert_consumer_handler_2(event: Event) -> Result[bool, Error]:
        spy_consumer_handler_2.append(event)
        result = simulated_results_handler_2.pop(0)
        return result

    event = EventUserCreatedMother.random()
    subscribers = [
        EventSubscriber(
            event_name=event.event_name,
            event_version=event.event_version,
            handlers=[assert_consumer_handler_1, assert_consumer_handler_2],
        )
    ]

    configurer = RabbitMqEventConfigurerMother.with_retry_ttl_10ms()
    configurer.configure_subscribers(subscribers)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    consumer = RabbitMqEventConsumerMother.with_max_retries(max_retries_allowed)
    consumer.add_subscribers(subscribers)
    consumer.add_handler_on_store(assert_consumer_event_store)

    consumer.start()

    sleep(1.0)

    consumer.stop()
    configurer.clear()

    spy_consumer_event_store.assert_number_unique_events(1)
    spy_consumer_event_store.assert_first_event(event)
    spy_consumer_event_store.assert_count_by_event_id(
        event.event_id, expected_number_event_consumed_by_store
    )

    spy_consumer_handler_1.assert_number_unique_events(1)
    spy_consumer_handler_1.assert_first_event(event)
    spy_consumer_handler_1.assert_count_by_event_id(
        event.event_id, expected_number_event_consumed_by_handler_1
    )

    spy_consumer_handler_2.assert_number_unique_events(1)
    spy_consumer_handler_2.assert_first_event(event)
    spy_consumer_handler_2.assert_count_by_event_id(
        event.event_id, expected_number_event_consumed_by_handler_2
    )
